#!/usr/bin/env python
# MAPIES base

from functools import wraps
from datetime import datetime, timedelta
import sys
from mapies.util.func_tools import timeit, time_domain_selection, frequency_int, error_estimation
from mapies.grids.monarch import RotatedGrid, IrregularRotatedGrid, RegularGrid
import time
import logging
import numpy as np
import pandas as pd
import xarray as xr
#import matplotlib
#matplotlib.use("TkAgg") # Use tinker to perform plt.show() as it is GUI supported 
import matplotlib.pyplot as plt
import cartopy
import cartopy.crs as ccrs
import os
import yaml
import matplotlib.cm as cm
from matplotlib.colors import ListedColormap
import matplotlib.colors as mcolors

#logger = logging.getLogger(__name__)
#logging.basicConfig(filename='../logs/mapies.log', level=logging.DEBUG)

class MAPIES:
    """
    Base class for only base functions used in every Mapies run
    
    For example Reading and writing functions
    """
    def __init__(self, start_date, end_date, **kwargs):
        """
        INITIATE the MAPIES class with the variables used in every run
        """
        self.np_typetime = "datetime64[s]"
        self.int_type = 'int64'
        self.time_orig = np.datetime64("1900-01-01T00:00:00")
        self.start_date = start_date
        self.end_date = end_date
        self.apply_qa = True
        self.geospatial_crop = None

        self.grid_dict = {
            "rotated":RotatedGrid, 
            "regular":RegularGrid, 
            "ireg_rotated":IrregularRotatedGrid
        }


    def read_config(self, **kwargs):
        """
        Read yaml config file
        """
        module_dir = os.path.dirname(__file__)


        # Let the user read in their own config
        config_file = kwargs.get("config_file")
        if config_file:
            try:
                self.config = yaml.safe_load(open(config_file))
            except FileNotFoundError:
                #logger.error("Config file passed does not exist: Using default config instead")
                self.config = yaml.safe_load(open(os.path.join(module_dir, "config/satellite_config.yaml")))
        else:
            self.config = yaml.safe_load(open(os.path.join(module_dir, "config/satellite_config.yaml")))


        variable_dict = self.config[self.datatype]["variables"]
        self.time_var = variable_dict["time_variable"]
        self.lon_var = variable_dict["lon_variable"]
        self.lat_var = variable_dict["lat_variable"]

        # If obs_var passed then run data analysis on that variable if not pull default
        obs_var = kwargs.get("obs_var")
        if obs_var:
            self.obs_var = obs_var
        else:
            self.obs_var = variable_dict["obs_variable"]
        

        grid_dict = self.config[self.datatype]["grids"]
        grid_repr = kwargs.get("grid_repr")
        if grid_repr:
            self.grid_repr = grid_repr
        else:
            self.grid_repr = grid_dict["grid_repr"] #str

        if self.grid_repr:
            # Cread grid object if grid repr is selected in the input
            self.grid_config = grid_dict[self.grid_repr]
            self.grid = self.grid_dict[self.grid_repr](**self.grid_config)
            print(self.grid)
        

        qa_dict = self.config[self.datatype]["qa"]
        
        self.qualityFlags = qa_dict["qa_variables"]



    # TODO this can be moved to a plotting function python file
    @timeit
    def plot_2D_obs_custom(self, lon_values=None, lat_values=None, obs_values=None, outdir=None, title=None, filename=None, display_fig=False):
        """
        General method for plotting observations with CB-RdYlBu colormap and arrow-like ends for out-of-bound values.
        """

        # Set Basemap projection
        figsize = (15, 10)
        markersize = 2.5
        proj = ccrs.PlateCarree()

        # Define discrete levels for the color bar
        if self.datatype == "viirs":
            label = "AOD"
            levels = [0.02, 0.04, 0.06, 0.08, 0.10, 0.15, 0.20, 0.30, 0.40, 0.50, 0.60, 0.70, 0.80, 1.00]
            level_format = [f"{level:.2f}" for level in levels]
        elif self.datatype == "tropomi":
            #TODO remove hard coding
            label = "NO2"
            levels = [0.00001, 0.00002, 0.00003, 0.00004, 0.00005, 0.00006, 0.00007, 0.00008, 0.00009, 0.00010, 0.00015, 0.00020, 0.00030, 0.00040]
            level_format = [f"{level:.5f}" for level in levels]
        # Define the custom color list
        colors = [
            '#290AD8', '#2645FA', '#3A91FF', '#66CBFF', '#99EEFF', 
            '#CDFCFF', '#F2FFD9', '#FFEEAA', '#FFCA88', '#FC966A', 
            '#ED5750', '#CB1C2D', '#9A001E', '#68001D', '#690067'
        ]

        # Create the custom colormap
        cmap = ListedColormap(colors)
        norm = mcolors.BoundaryNorm(boundaries=levels, ncolors=len(colors), extend="both")

        # Create the plot and add features
        fig, ax = plt.subplots(subplot_kw={"projection": proj}, figsize=figsize)
        ax.gridlines()
        ax.coastlines(resolution="10m")

        print("Plotting observations (new method)")

        # Scatter plot with discrete color mapping
        im = ax.scatter(
            lon_values, lat_values, markersize, c=obs_values, cmap=cmap, norm=norm, transform=proj
        )

        # Add the color bar
        cbar = fig.colorbar(im, ax=ax, orientation="vertical", pad=0.05, ticks=levels, extend="both")
        cbar.set_label(label)
        cbar.ax.set_yticklabels(level_format)  # Format tick labels

        # Add title and save the plot
        ax.set_title(title)

        # Ensure the output directory exists
        #os.makedirs(outdir, exist_ok=True)

        filepath = os.path.join(outdir, filename)
        plt.savefig(filepath, format="png")
        if display_fig:
            plt.show()
        print(f"Saved plot: {filepath}")
        

    # TODO this can be moved to a plotting function python file
    @timeit 
    def plot_2D_num_obs_custom(
        self, lon_values=None, 
        lat_values=None, 
        num_obs=None, 
        outdir="./",
        title=None, 
        filename=None,
        display_fig=False
        ):
        """
        General method for plotting the number of observations contributing to the final averages.
        """

        # Set Basemap projection
        figsize = (15, 10)
        markersize = 2.5
        proj = ccrs.PlateCarree()

        # Define discrete levels for the color bar
        unique_values = np.unique(num_obs)  # Unique values in num_obs
        levels = np.linspace(unique_values.min(), unique_values.max(), unique_values.max()-unique_values.min()+1)


        # Use a discrete colormap

        cmap = plt.cm.rainbow
        norm = mcolors.Normalize(vmin=unique_values.min(), vmax=unique_values.max())
        # norm = mcolors.BoundaryNorm(np.arange(unique_values.min()-0.5, unique_values.max()+1, 1), cmap.N)

        # Create the plot and add features
        fig, ax = plt.subplots(subplot_kw={"projection": proj}, figsize=figsize)
        ax.gridlines(draw_labels=True)
        ax.coastlines(resolution="10m")

        # Scatter plot for the number of observations
        im = ax.scatter(
            lon_values, lat_values, markersize, c=num_obs, cmap=cmap, norm=norm, transform=proj
        )

        # Add the discrete colorbar without arrow-like ends
        cbar = fig.colorbar(im, ax=ax, orientation="vertical")
        cbar.set_label("Number of Observations")
        # cbar.ax.set_yticklabels([f"{level}" for level in range(unique_values)]) 

        # Add title and save the plot
        ax.set_title(title or "Number of Observations Plot")

        # Ensure the output directory exists
        os.makedirs(outdir, exist_ok=True)

        filepath = os.path.join(self.dest, filename or "num_observation_2D_plot.png")
        plt.savefig(filepath, format="png")
        if display_fig:
            plt.show()
        print(f"Saved plot: {filepath}")
        
        #return fig


    @staticmethod
    def to_xarray(coords:dict, data_vars:dict, **kwargs):
        """
        Method to convert numpy arrays to xarray opject
        """
        attrs = kwargs.get("attrs")
        if attrs is None:
            attrs=dict()

        ds = xr.Dataset(data_vars=data_vars, coords=coords, attrs=attrs)
        return ds
